Unit IVoice1;

Interface

Uses IntvSound;

{
 * ============================================================================
 *  Title:    IntellIVoice Emulation
 *  Author:   J. Zbiciak
 *  $Id: IVoice.h,v 1.1 2000/09/28 00:05:49 im14u2c Exp $
 *
 * Converted to Pascal by John Dullea
 * ============================================================================
}

Type
  PROMPage = ^TROMPage;
  TROMPage = Packed Array[0..4095] Of Byte;

  TLPC12 = Record
    Rpt, Cnt : Integer;                     { Repeat counter, Period down-counter.   }
    Per      : Integer;                     { Period                                 }
    Rng      : LongInt;                     { Random Number Generator                }
    Amp      : Integer;                     { Amplitude                              }
    F_Coef   : Array[0..5] Of Integer;      { F0 through F5.                         }
    B_Coef   : Array[0..5] Of Integer;      { B0 through B5.                         }
    Z_Data   : Array[0..5,0..1] Of Integer; { Time-delay data for the filter stages. }
    R        : Array[0..15] Of Byte;        { The encoded register set.              }
    Interp   : Boolean;
  End;

  TIVoice = Record
    Rate          : Word;       { Sample rate, Window size.                    }
    Filt          : TLPC12;     { 12-pole filter                               }
    LRQ           : Word;       { Load ReQuest.  == 0 if we can accept a load  }
    ALD           : Integer;    { Address LoaD.  < 0 if no command pending.    }
    PC            : Word;       { Microcontroller's PC value.                  }
    Stack         : Word;       { Microcontroller's PC stack.                  }
    Fifo_Sel      : Boolean;    { True when executing from FIFO.               }
    Halted        : Boolean;    { True when CPU is Halted.                     }
    Mode          : LongInt;    { Mode register.                               }
    Page          : Cardinal;   { Page set by SETPAGE                          }
    Fifo_Head     : LongInt;    { FIFO head pointer (where new data goes).     }
    Fifo_Tail     : LongInt;    { FIFO tail pointer (where data comes from).   }
    Fifo_BitP     : LongInt;    { FIFO bit-pointer (for partial decles).       }
    Fifo          : Array[0..63] Of Word;     { The 64-decle FIFO.      }
    Rom           : Array[0..15] Of PROMPage; { 4K ROM pages.           }
  End;

{ ======================================================================== }
{  This program is free software; you can redistribute it and/or modify    }
{  it under the terms of the GNU General Public License as published by    }
{  the Free Software Foundation; either version 2 of the License, or       }
{  (at your option) any later version.                                     }
{                                                                          }
{  This program is distributed in the hope that it will be useful,         }
{  but WITHOUT ANY WARRANTY; without even the implied warranty of          }
{  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       }
{  General Public License for more details.                                }
{                                                                          }
{  You should have received a copy of the GNU General Public License       }
{  along with this program; if not, write to the Free Software             }
{  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               }
{ ======================================================================== }
{                 Copyright (c) 1998-2000, Joseph Zbiciak                  }
{ ======================================================================== }


{
 * ============================================================================
 *  Title:    IntellIVoice Emulation
 *  Author:   J. Zbiciak
 *  $Id: IVoice.c,v 1.1 2000/09/28 00:05:49 im14u2c Exp $
 * ============================================================================
 *  This module actually attempts to emulate the IntellIVoice.  Wild!
 * ============================================================================
 *  The IntellIVoice is mapped into two locations in memory, $0080-$0081.
 *  (This ignores the separate 8-bit bus that the SPB-640 provides, since
 *  nothing uses it and I haven't emulated it.)
 *
 *  Location $0080 provides an interface to the "Address LoaD" (ALD)
 *  mechanism on the SP0256.  Reads from this address return the current
 *  "Load ReQuest" (LRQ) state in bit 15.  When LRQ is 1 (ie. bit 15 of
 *  location $0080 reads as 1), the SP0256 is ready to receive a new command.
 *  A new command address may then be written to location $0080 to trigger
 *  the playback of a sound.  Note that command address register is actually
 *  a 1-deep FIFO, and so LRQ with go to 1 before the SP0256 is finished
 *  speaking.
 *
 *  Location $0081 provides an interface to the SPB-640's 64-decle speech
 *  FIFO.  Reads from this address return the "FIFO full" state in bit 15.
 *  When bit 15 reads as 0, the FIFO has room for at least 1 more decle.
 *  Writes to this address can either clear the FIFO, or provide new data
 *  to the FIFO.  To clear the FIFO, write a value with Bit 10 == 1.
 *  To put a decle into the FIFO, write a value with Bit 10 == 0.  It's
 *  currently unknown what happens when a program attempts to write to the
 *  FIFO when the FIFO is full.  This emulation drops the extra data.
 *
 *  The exact format of the SP0256 speech data, as well as the overall
 *  system view from the SP0256's perspective is documented elsewhere.
 * ============================================================================
 }

Const
  SCBUF_SIZE = 4096;             { Must be power of 2               }
  SCBUF_MASK = SCBUF_SIZE - 1;
  PER_PAUSE  = 64;               { Equiv timing period for pauses.  }
  PER_NOISE  = 64;               { Equiv timing period for noise.   }

  FIFO_ADDR  = $C000;            { SP0256 address of speech FIFO.   }

  Stage_Map: Array[0..5] Of Integer = (0, 1, 2, 3, 4, 5);

{ ======================================================================== }
{  IVoice_QTBL  -- Coefficient Quantization Table.  This comes from a      }
{                  SP0250 data sheet, and should be correct for SP0256.    }
{ ======================================================================== }
  QTbl : Array[0..127] Of Integer =
   (0,      9,      12,     25,     33,     41,     49,     57,
    65,     73,     81,     89,     97,     105,    113,    121,
    129,    137,    145,    153,    161,    169,    177,    185,
    193,    201,    209,    217,    225,    233,    241,    249,
    257,    265,    273,    281,    289,    297,    301,    305,
    309,    313,    317,    321,    325,    329,    333,    337,
    341,    345,    349,    353,    357,    361,    365,    369,
    373,    377,    381,    385,    389,    393,    397,    401,
    405,    409,    413,    417,    421,    425,    427,    429,
    431,    433,    435,    437,    439,    441,    443,    445,
    447,    449,    451,    453,    455,    457,    459,    461,
    463,    465,    467,    469,    471,    473,    475,    477,
    479,    481,    482,    483,    484,    485,    486,    487,
    488,    489,    490,    491,    492,    493,    494,    495,
    496,    497,    498,    499,    500,    501,    502,    503,
    504,    505,    506,    507,    508,    509,    510,    511);

Var
  Voice : TIVoice;
  Mask  : Packed Array[0..4095] Of Byte;

Function  IVoice_Tk(Var IVoice: TIVoice; Len: Word): LongInt;
Function  IVoice_Rd(Var IVoice: TIVoice; Addr: Word): Word;
Procedure IVoice_Wr(Var IVoice: TIVoice; Addr,Data: Word);
Function  IVoice_Init(Var IVoice: TIVoice; Rate: Word): Boolean;

Implementation

Const
  Samp_Count : Word    = 0;
  Samp       : Integer = 0;

Function Limit(S: Integer): Integer;
{ ======================================================================== }
{  LIMIT            -- Limiter function for digital sample output.         }
{ ======================================================================== }
Begin
  If S >  127 Then S :=  127;
  If S < -128 Then S := -128;
  Limit := S;
End; { Limit }

Function LPC12_Update(Var F: TLPC12; Num_Samp: Integer;
                      Index: Word; Rate: Word): Word;
{ ======================================================================== }
{  LPC12_Update -- Update the 12-pole filter, outputting samples.          }
{ ======================================================================== }
Var I: Integer;

  Function GenerateNewSample: Boolean;
  Var
    J      : Integer;
    Do_Int : Boolean;

  Begin
    { ---------------------------------------------------------------- }
    {  Generate a series of periodic impulses, or random noise.        }
    { ---------------------------------------------------------------- }

    Do_Int := False;
    If F.Per <> 0 Then
    Begin
      If F.Cnt <= 0 Then
      Begin
        Inc(F.Cnt,F.Per);
        Samp := F.Amp;
        Dec(F.Rpt);
        Do_Int := F.Interp;
        FillChar(F.Z_Data,SizeOf(F.Z_Data),#0);
      End
      Else
      Begin
        Samp := 0;
        Dec(F.Cnt);
      End;
    End
    Else
    Begin
      Dec(F.Cnt);
      If F.Cnt <= 0 Then
      Begin
        Do_Int := F.Interp;
        F.Cnt  := PER_NOISE;
        Dec(F.Rpt);
        FillChar(F.Z_Data,SizeOf(F.Z_Data),#0);
      End;
      Samp := -F.Amp;
      If (F.Rng And 1) <> 0 Then
      Begin
        F.Rng := F.Rng Xor $28000;
        Samp  := -Samp;
      End;
      F.Rng := F.Rng Shr 1;
    End;

    { ---------------------------------------------------------------- }
    {  If we need to, process the interpolation registers.             }
    { ---------------------------------------------------------------- }

    If Do_Int Then
    Begin
      Inc(F.R[0],F.R[14]);
      Inc(F.R[1],F.R[15]);
      F.Amp := (F.R[0] And $1F) Shl ((F.R[0] And $E0) Shr 5);
      F.Per := F.R[1];
    End;

    { ---------------------------------------------------------------- }
    {  Stop if we expire our repeat counter and return the actual      }
    {  number of samples we did.                                       }
    { ---------------------------------------------------------------- }
    If F.Rpt <= 0 Then
    Begin
      GenerateNewSample := False;
      Exit;
    End;

    { ---------------------------------------------------------------- }
    {  Each 2nd order stage looks like one of these.  The App. Manual  }
    {  gives the first form, the patent gives the second form.         }
    {  They're equivalent except for time delay.  I implement the      }
    {  first form.   (Note: 1/Z == 1 unit of time delay.)              }
    {                                                                  }
    {          ---->(+)-------->(+)----------+------>                  }
    {                ^           ^           |                         }
    {                |           |           |                         }
    {                |           |           |                         }
    {               [B]        [2*F]         |                         }
    {                ^           ^           |                         }
    {                |           |           |                         }
    {                |           |           |                         }
    {                +---[1/Z]<--+---[1/Z]<--+                         }
    {                                                                  }
    {                                                                  }
    {                +---[2*F]<---+                                    }
    {                |            |                                    }
    {                |            |                                    }
    {                v            |                                    }
    {          ---->(+)-->[1/Z]-->+-->[1/Z]---+------>                 }
    {                ^                        |                        }
    {                |                        |                        }
    {                |                        |                        }
    {                +-----------[B]<---------+                        }
    {                                                                  }
    { ---------------------------------------------------------------- }
    For J := 0 To 5 Do
    Begin
      Inc(Samp,((F.B_Coef[J] * F.Z_Data[J,1]) Div 512));
      Inc(Samp,((F.F_Coef[J] * F.Z_Data[J,0]) Div 256));
      F.Z_Data[J,1] := F.Z_Data[J,0];
      F.Z_Data[J,0] := Samp;
    End; { For J }

    { Write the sample, adjusting for the playback rate }

    Samp := Limit(Samp Div 16) + 128;

    GenerateNewSample := True;
  End; // GenerateNewSample

Label L1;
Begin
  I := 0;

  { -------------------------------------------------------------------- }
  { If there is a leftover sample, handle it first.                      }
  { -------------------------------------------------------------------- }
  While (Samp_Count < Rate) And (I < Num_Samp) Do
  Begin
    VoiceBuf[Index And SCBUF_MASK] := Byte(Samp);
    Inc(Index);
    Inc(Samp_Count,10000);
    Inc(I);
  End; { While }

  { -------------------------------------------------------------------- }
  {  Iterate up to the desired number of samples.  We actually may       }
  {  break out early if our repeat count expires.                        }
  { -------------------------------------------------------------------- }

  While I < Num_Samp Do
  Begin
    If Not GenerateNewSample Then Goto L1;

    If Samp_Count >= Rate Then Dec(Samp_Count,Rate);
    While (Samp_Count < Rate) And (I < Num_Samp) Do
    Begin
      VoiceBuf[Index And SCBUF_MASK] := Byte(Samp);
      Inc(Index);
      Inc(Samp_Count,10000);
      Inc(I);
    End; { While }
  End; { While }
L1:
  LPC12_Update := I;
End; { LPC12_Update }

Procedure LPC12_RegDec(Var F: TLPC12);
{ ======================================================================== }
{  LPC12_RegDec -- Decode the register set in the filter bank.             }
{ ======================================================================== }
Var I: Integer;

  Function IQ(X1: Byte): Integer;
  Var X: ShortInt;
  Begin
    X := ShortInt(X1);
    If X < 0
     Then IQ := QTbl[$7F And -X]
     Else IQ := -QTbl[X];
  End; { IQ }

Begin

  { -------------------------------------------------------------------- }
  {  Decode the Amplitude and Period registers.  Force the 'cnt' to 0    }
  {  to get an initial impulse.  We compensate elsewhere by setting      }
  {  the repeat count to "repeat + 1".                                   }
  { -------------------------------------------------------------------- }
  F.Amp := (F.R[0] And $1F) Shl ((F.R[0] And $E0) Shr 5);
  F.Cnt := 0;
  F.Per := F.R[1];

  { -------------------------------------------------------------------- }
  {  Decode the filter coefficients from the quant table.                }
  { -------------------------------------------------------------------- }
  For I := 0 To 5 Do
  Begin
    F.B_Coef[Stage_Map[I]] := IQ(F.R[2 + 2 * I]);
    F.F_Coef[Stage_Map[I]] := IQ(F.R[3 + 2 * I]);
  End; { For I }

  { -------------------------------------------------------------------- }
  {  Set the Interp flag based on whether we have interpolation parms    }
  { -------------------------------------------------------------------- }
  F.Interp := ((F.R[14] Or F.R[15]) <> 0);
End; { LPC12_RegDec }

{ ======================================================================== }
{  SP0256_DATAFMT   -- Data format table for the SP0256's microcontroller  }
{                                                                          }
{  len     4 bits      Length of field to extract                          }
{  lshift  4 bits      Left-shift amount on field                          }
{  param   4 bits      Parameter number being updated                      }
{  delta   1 bit       This is a delta-update.  (Implies sign-extend)      }
{  field   1 bit       This is a field replace.                            }
{  clr5    1 bit       Clear F5, B5.                                       }
{  clrall  1 bit       Clear all before doing this update                  }
{ ======================================================================== }

Const
  SP0256_DataFmt : Array[0..176] of Word =
   ($8000,
    $8008,$0108,$0208,$0308,$0408,$0508,$0608,$0708,$0808,$0908,$0A08,$0B08,$0C08,$0D08,$0E08,$0F08,
    $8026,$0108,$0834,$0926,$0A17,$0B26,$0C08,$0D08,$8026,$0108,$0816,$0917,$0A08,$0B08,$0C08,$0D08,
    $4000,$0026,$2926,$2B26,$2D08,$4000,$0026,$2917,$2B08,$2D08,$0000,$0000,
    $1024,$1105,$1243,$1333,$1443,$1533,$1643,$1733,$1833,$1924,$1A14,$1B24,$1C05,$1D05,$1024,$1105,
    $1214,$1324,$1414,$1524,$1614,$1724,$1814,$1915,$1A05,$1B05,$1C05,$1D05,
    $4000,$0026,$2335,$2535,$2735,$4000,$0026,$2326,$2526,$2726,
    $8026,$0108,$0243,$0335,$0443,$0535,$0643,$0735,$0834,$0926,$0A17,$0B26,$0E05,$0F05,$8026,$0108,
    $0216,$0326,$0416,$0526,$0616,$0726,$0816,$0917,$0A08,$0B08,$0E05,$0F05,
    $1024,$1105,$1833,$1924,$1A14,$1B24,$1C05,$1D05,$1024,$1105,$1814,$1915,$1A05,$1B05,$1C05,$1D05,
    $0026,$0108,
    $8026,$0108,$0243,$0335,$0443,$0535,$0643,$0735,$0834,$0926,$0A17,$0B26,$0C08,$0D08,$0E05,$0F05,
    $8026,$0108,$0216,$0326,$0416,$0526,$0616,$0726,$0816,$0917,$0A08,$0B08,$0C08,$0D08,$0E05,$0F05,
    $4000,$0026,$0108,$2335,$2535,$2735,$0E05,$0F05,$4000,$0026,$0108,$2326,$2526,$2726,$0E05,$0F05);

Function CR(l,s,p,d,f,c5,ca: Word): Word;
Begin
  CR := (((l)  And 15) Shl  0) Or
        (((s)  And 15) Shl  4) Or
        (((p)  And 15) Shl  8) Or
        (((d)  And  1) Shl 12) Or
        (((f)  And  1) Shl 13) Or
        (((c5) And  1) Shl 14) Or
        (((ca) And  1) Shl 15);
End; { CR }

Const
  CR_DELTA = $1000;
  CR_FIELD = $2000;
  CR_CLR5  = $4000;
  CR_CLRA  = $8000;

Function CR_LEN(X: Word): Word;
Begin
  CR_LEN := X And $F;
End; { CR_LEN }

Function CR_SHF(X: Word): Word;
Begin
  CR_SHF := (X Shr 4) And 15;
End; { CR_SHF }

Function CR_PRM(X: Word): Word;
Begin
  CR_PRM := (X Shr 8) And 15;
End; { CR_PRM }

Type
  TOperand = (AM, PR, B0, F0, B1, F1, B2, F2, B3, F3, B4, F4, B5, F5, IA, IP);

(*
enum { AM = 0, PR, B0, F0, B1, F1, B2, F2, B3, F3, B4, F4, B5, F5, IA, IP };

static const uint_16 sp0256_datafmt[] =
{
    { -------------------------------------------------------------------- }
    {  OPCODE 1111: PAUSE                                                  }
    { -------------------------------------------------------------------- }
    {    0 }  CR( 0,  0,  0,  0,  0,  0,  1),     {  Clear all   }

    { -------------------------------------------------------------------- }
    {  Opcode 0001: LOADALL                                                }
    { -------------------------------------------------------------------- }
                { All modes                }
    {    1 }  CR( 8,  0,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {    2 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {    3 }  CR( 8,  0,  B0, 0,  0,  0,  0),     {  B0          }
    {    4 }  CR( 8,  0,  F0, 0,  0,  0,  0),     {  F0          }
    {    5 }  CR( 8,  0,  B1, 0,  0,  0,  0),     {  B1          }
    {    6 }  CR( 8,  0,  F1, 0,  0,  0,  0),     {  F1          }
    {    7 }  CR( 8,  0,  B2, 0,  0,  0,  0),     {  B2          }
    {    8 }  CR( 8,  0,  F2, 0,  0,  0,  0),     {  F2          }
    {    9 }  CR( 8,  0,  B3, 0,  0,  0,  0),     {  B3          }
    {   10 }  CR( 8,  0,  F3, 0,  0,  0,  0),     {  F3          }
    {   11 }  CR( 8,  0,  B4, 0,  0,  0,  0),     {  B4          }
    {   12 }  CR( 8,  0,  F4, 0,  0,  0,  0),     {  F4          }
    {   13 }  CR( 8,  0,  B5, 0,  0,  0,  0),     {  B5          }
    {   14 }  CR( 8,  0,  F5, 0,  0,  0,  0),     {  F5          }
                { Mode 01 and 11 only      }
    {   15 }  CR( 8,  0,  IA, 0,  0,  0,  0),     {  Amp Interp  }
    {   16 }  CR( 8,  0,  IP, 0,  0,  0,  0),     {  Pit Interp  }

    { -------------------------------------------------------------------- }
    {  Opcode 0100: LOAD_4                                                 }
    { -------------------------------------------------------------------- }
                { Mode 00 and 01           }
    {   17 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {   18 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {   19 }  CR( 4,  3,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {   20 }  CR( 6,  2,  F3, 0,  0,  0,  0),     {  F3          }
    {   21 }  CR( 7,  1,  B4, 0,  0,  0,  0),     {  B4          }
    {   22 }  CR( 6,  2,  F4, 0,  0,  0,  0),     {  F4          }
                { Mode 01 only             }
    {   23 }  CR( 8,  0,  B5, 0,  0,  0,  0),     {  B5          }
    {   24 }  CR( 8,  0,  F5, 0,  0,  0,  0),     {  F5          }

                { Mode 10 and 11           }
    {   25 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {   26 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {   27 }  CR( 6,  1,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {   28 }  CR( 7,  1,  F3, 0,  0,  0,  0),     {  F3          }
    {   29 }  CR( 8,  0,  B4, 0,  0,  0,  0),     {  B4          }
    {   30 }  CR( 8,  0,  F4, 0,  0,  0,  0),     {  F4          }
                { Mode 11 only             }
    {   31 }  CR( 8,  0,  B5, 0,  0,  0,  0),     {  B5          }
    {   32 }  CR( 8,  0,  F5, 0,  0,  0,  0),     {  F5          }

    { -------------------------------------------------------------------- }
    {  Opcode 0110: SETMSB_6                                               }
    { -------------------------------------------------------------------- }
                { Mode 00 only             }
    {   33 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 00 and 01           }
    {   34 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {   35 }  CR( 6,  2,  F3, 0,  1,  0,  0),     {  F3 (5 MSBs) }
    {   36 }  CR( 6,  2,  F4, 0,  1,  0,  0),     {  F4 (5 MSBs) }
                { Mode 01 only             }
    {   37 }  CR( 8,  0,  F5, 0,  1,  0,  0),     {  F5 (5 MSBs) }

                { Mode 10 only             }
    {   38 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 10 and 11           }
    {   39 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {   40 }  CR( 7,  1,  F3, 0,  1,  0,  0),     {  F3 (6 MSBs) }
    {   41 }  CR( 8,  0,  F4, 0,  1,  0,  0),     {  F4 (6 MSBs) }
                { Mode 11 only             }
    {   42 }  CR( 8,  0,  F5, 0,  1,  0,  0),     {  F5 (6 MSBs) }

    {   43 }  0,  { unused }
    {   44 }  0,  { unused }

    { -------------------------------------------------------------------- }
    {  Opcode 1001: DELTA_9                                                }
    { -------------------------------------------------------------------- }
                { Mode 00 and 01           }
    {   45 }  CR( 4,  2,  AM, 1,  0,  0,  0),     {  Amplitude   }
    {   46 }  CR( 5,  0,  PR, 1,  0,  0,  0),     {  Period      }
    {   47 }  CR( 3,  4,  B0, 1,  0,  0,  0),     {  B0 4 MSBs   }
    {   48 }  CR( 3,  3,  F0, 1,  0,  0,  0),     {  F0 5 MSBs   }
    {   49 }  CR( 3,  4,  B1, 1,  0,  0,  0),     {  B1 4 MSBs   }
    {   50 }  CR( 3,  3,  F1, 1,  0,  0,  0),     {  F1 5 MSBs   }
    {   51 }  CR( 3,  4,  B2, 1,  0,  0,  0),     {  B2 4 MSBs   }
    {   52 }  CR( 3,  3,  F2, 1,  0,  0,  0),     {  F2 5 MSBs   }
    {   53 }  CR( 3,  3,  B3, 1,  0,  0,  0),     {  B3 5 MSBs   }
    {   54 }  CR( 4,  2,  F3, 1,  0,  0,  0),     {  F3 6 MSBs   }
    {   55 }  CR( 4,  1,  B4, 1,  0,  0,  0),     {  B4 7 MSBs   }
    {   56 }  CR( 4,  2,  F4, 1,  0,  0,  0),     {  F4 6 MSBs   }
                { Mode 01 only             }
    {   57 }  CR( 5,  0,  B5, 1,  0,  0,  0),     {  B5 8 MSBs   }
    {   58 }  CR( 5,  0,  F5, 1,  0,  0,  0),     {  F5 8 MSBs   }

                { Mode 10 and 11           }
    {   59 }  CR( 4,  2,  AM, 1,  0,  0,  0),     {  Amplitude   }
    {   60 }  CR( 5,  0,  PR, 1,  0,  0,  0),     {  Period      }
    {   61 }  CR( 4,  1,  B0, 1,  0,  0,  0),     {  B0 7 MSBs   }
    {   62 }  CR( 4,  2,  F0, 1,  0,  0,  0),     {  F0 6 MSBs   }
    {   63 }  CR( 4,  1,  B1, 1,  0,  0,  0),     {  B1 7 MSBs   }
    {   64 }  CR( 4,  2,  F1, 1,  0,  0,  0),     {  F1 6 MSBs   }
    {   65 }  CR( 4,  1,  B2, 1,  0,  0,  0),     {  B2 7 MSBs   }
    {   66 }  CR( 4,  2,  F2, 1,  0,  0,  0),     {  F2 6 MSBs   }
    {   67 }  CR( 4,  1,  B3, 1,  0,  0,  0),     {  B3 7 MSBs   }
    {   68 }  CR( 5,  1,  F3, 1,  0,  0,  0),     {  F3 7 MSBs   }
    {   69 }  CR( 5,  0,  B4, 1,  0,  0,  0),     {  B4 8 MSBs   }
    {   70 }  CR( 5,  0,  F4, 1,  0,  0,  0),     {  F4 8 MSBs   }
                { Mode 11 only             }
    {   71 }  CR( 5,  0,  B5, 1,  0,  0,  0),     {  B5 8 MSBs   }
    {   72 }  CR( 5,  0,  F5, 1,  0,  0,  0),     {  F5 8 MSBs   }

    { -------------------------------------------------------------------- }
    {  Opcode 1010: SETMSB_A                                               }
    { -------------------------------------------------------------------- }
                { Mode 00 only             }
    {   73 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 00 and 01           }
    {   74 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {   75 }  CR( 5,  3,  F0, 0,  1,  0,  0),     {  F0 (5 MSBs) }
    {   76 }  CR( 5,  3,  F1, 0,  1,  0,  0),     {  F1 (5 MSBs) }
    {   77 }  CR( 5,  3,  F2, 0,  1,  0,  0),     {  F2 (5 MSBs) }

                { Mode 10 only             }
    {   78 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 10 and 11           }
    {   79 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {   80 }  CR( 6,  2,  F0, 0,  1,  0,  0),     {  F0 (6 MSBs) }
    {   81 }  CR( 6,  2,  F1, 0,  1,  0,  0),     {  F1 (6 MSBs) }
    {   82 }  CR( 6,  2,  F2, 0,  1,  0,  0),     {  F2 (6 MSBs) }

    { -------------------------------------------------------------------- }
    {  Opcode 0010: LOAD_2  Mode 00 and 10                                 }
    {  Opcode 1100: LOAD_C  Mode 00 and 10                                 }
    { -------------------------------------------------------------------- }
                { LOAD_2, LOAD_C  Mode 00  }
    {   83 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {   84 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {   85 }  CR( 3,  4,  B0, 0,  0,  0,  0),     {  B0 (S=0)    }
    {   86 }  CR( 5,  3,  F0, 0,  0,  0,  0),     {  F0          }
    {   87 }  CR( 3,  4,  B1, 0,  0,  0,  0),     {  B1 (S=0)    }
    {   88 }  CR( 5,  3,  F1, 0,  0,  0,  0),     {  F1          }
    {   89 }  CR( 3,  4,  B2, 0,  0,  0,  0),     {  B2 (S=0)    }
    {   90 }  CR( 5,  3,  F2, 0,  0,  0,  0),     {  F2          }
    {   91 }  CR( 4,  3,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {   92 }  CR( 6,  2,  F3, 0,  0,  0,  0),     {  F3          }
    {   93 }  CR( 7,  1,  B4, 0,  0,  0,  0),     {  B4          }
    {   94 }  CR( 6,  2,  F4, 0,  0,  0,  0),     {  F4          }
                { LOAD_2 only              }
    {   95 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {   96 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }

                { LOAD_2, LOAD_C  Mode 10  }
    {   97 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {   98 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {   99 }  CR( 6,  1,  B0, 0,  0,  0,  0),     {  B0 (S=0)    }
    {  100 }  CR( 6,  2,  F0, 0,  0,  0,  0),     {  F0          }
    {  101 }  CR( 6,  1,  B1, 0,  0,  0,  0),     {  B1 (S=0)    }
    {  102 }  CR( 6,  2,  F1, 0,  0,  0,  0),     {  F1          }
    {  103 }  CR( 6,  1,  B2, 0,  0,  0,  0),     {  B2 (S=0)    }
    {  104 }  CR( 6,  2,  F2, 0,  0,  0,  0),     {  F2          }
    {  105 }  CR( 6,  1,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {  106 }  CR( 7,  1,  F3, 0,  0,  0,  0),     {  F3          }
    {  107 }  CR( 8,  0,  B4, 0,  0,  0,  0),     {  B4          }
    {  108 }  CR( 8,  0,  F4, 0,  0,  0,  0),     {  F4          }
                { LOAD_2 only              }
    {  109 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {  110 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }

    { -------------------------------------------------------------------- }
    {  OPCODE 1101: DELTA_D                                                }
    { -------------------------------------------------------------------- }
                { Mode 00 and 01           }
    {  111 }  CR( 4,  2,  AM, 1,  0,  0,  0),     {  Amplitude   }
    {  112 }  CR( 5,  0,  PR, 1,  0,  0,  0),     {  Period      }
    {  113 }  CR( 3,  3,  B3, 1,  0,  0,  0),     {  B3 5 MSBs   }
    {  114 }  CR( 4,  2,  F3, 1,  0,  0,  0),     {  F3 6 MSBs   }
    {  115 }  CR( 4,  1,  B4, 1,  0,  0,  0),     {  B4 7 MSBs   }
    {  116 }  CR( 4,  2,  F4, 1,  0,  0,  0),     {  F4 6 MSBs   }
                { Mode 01 only             }
    {  117 }  CR( 5,  0,  B5, 1,  0,  0,  0),     {  B5 8 MSBs   }
    {  118 }  CR( 5,  0,  F5, 1,  0,  0,  0),     {  F5 8 MSBs   }

                { Mode 10 and 11           }
    {  119 }  CR( 4,  2,  AM, 1,  0,  0,  0),     {  Amplitude   }
    {  120 }  CR( 5,  0,  PR, 1,  0,  0,  0),     {  Period      }
    {  121 }  CR( 4,  1,  B3, 1,  0,  0,  0),     {  B3 7 MSBs   }
    {  122 }  CR( 5,  1,  F3, 1,  0,  0,  0),     {  F3 7 MSBs   }
    {  123 }  CR( 5,  0,  B4, 1,  0,  0,  0),     {  B4 8 MSBs   }
    {  124 }  CR( 5,  0,  F4, 1,  0,  0,  0),     {  F4 8 MSBs   }
                { Mode 11 only             }
    {  125 }  CR( 5,  0,  B5, 1,  0,  0,  0),     {  B5 8 MSBs   }
    {  126 }  CR( 5,  0,  F5, 1,  0,  0,  0),     {  F5 8 MSBs   }

    { -------------------------------------------------------------------- }
    {  OPCODE 1110: LOAD_E                                                 }
    { -------------------------------------------------------------------- }
    {  127 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {  128 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }

    { -------------------------------------------------------------------- }
    {  Opcode 0010: LOAD_2  Mode 01 and 11                                 }
    {  Opcode 1100: LOAD_C  Mode 01 and 11                                 }
    { -------------------------------------------------------------------- }
                { LOAD_2, LOAD_C  Mode 01  }
    {  129 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {  130 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {  131 }  CR( 3,  4,  B0, 0,  0,  0,  0),     {  B0 (S=0)    }
    {  132 }  CR( 5,  3,  F0, 0,  0,  0,  0),     {  F0          }
    {  133 }  CR( 3,  4,  B1, 0,  0,  0,  0),     {  B1 (S=0)    }
    {  134 }  CR( 5,  3,  F1, 0,  0,  0,  0),     {  F1          }
    {  135 }  CR( 3,  4,  B2, 0,  0,  0,  0),     {  B2 (S=0)    }
    {  136 }  CR( 5,  3,  F2, 0,  0,  0,  0),     {  F2          }
    {  137 }  CR( 4,  3,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {  138 }  CR( 6,  2,  F3, 0,  0,  0,  0),     {  F3          }
    {  139 }  CR( 7,  1,  B4, 0,  0,  0,  0),     {  B4          }
    {  140 }  CR( 6,  2,  F4, 0,  0,  0,  0),     {  F4          }
    {  141 }  CR( 8,  0,  B5, 0,  0,  0,  0),     {  B5          }
    {  142 }  CR( 8,  0,  F5, 0,  0,  0,  0),     {  F5          }
                { LOAD_2 only              }
    {  143 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {  144 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }

                { LOAD_2, LOAD_C  Mode 11  }
    {  145 }  CR( 6,  2,  AM, 0,  0,  0,  1),     {  Amplitude   }
    {  146 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {  147 }  CR( 6,  1,  B0, 0,  0,  0,  0),     {  B0 (S=0)    }
    {  148 }  CR( 6,  2,  F0, 0,  0,  0,  0),     {  F0          }
    {  149 }  CR( 6,  1,  B1, 0,  0,  0,  0),     {  B1 (S=0)    }
    {  150 }  CR( 6,  2,  F1, 0,  0,  0,  0),     {  F1          }
    {  151 }  CR( 6,  1,  B2, 0,  0,  0,  0),     {  B2 (S=0)    }
    {  152 }  CR( 6,  2,  F2, 0,  0,  0,  0),     {  F2          }
    {  153 }  CR( 6,  1,  B3, 0,  0,  0,  0),     {  B3 (S=0)    }
    {  154 }  CR( 7,  1,  F3, 0,  0,  0,  0),     {  F3          }
    {  155 }  CR( 8,  0,  B4, 0,  0,  0,  0),     {  B4          }
    {  156 }  CR( 8,  0,  F4, 0,  0,  0,  0),     {  F4          }
    {  157 }  CR( 8,  0,  B5, 0,  0,  0,  0),     {  B5          }
    {  158 }  CR( 8,  0,  F5, 0,  0,  0,  0),     {  F5          }
                { LOAD_2 only              }
    {  159 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {  160 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }

    { -------------------------------------------------------------------- }
    {  Opcode 0011: SETMSB_3                                               }
    {  Opcode 0101: SETMSB_5                                               }
    { -------------------------------------------------------------------- }
                { Mode 00 only             }
    {  161 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 00 and 01           }
    {  162 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {  163 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {  164 }  CR( 5,  3,  F0, 0,  1,  0,  0),     {  F0 (5 MSBs) }
    {  165 }  CR( 5,  3,  F1, 0,  1,  0,  0),     {  F1 (5 MSBs) }
    {  166 }  CR( 5,  3,  F2, 0,  1,  0,  0),     {  F2 (5 MSBs) }
                { SETMSB_3 only            }
    {  167 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {  168 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }

                { Mode 10 only             }
    {  169 }  CR( 0,  0,  0,  0,  0,  1,  0),     {  Clear 5     }
                { Mode 10 and 11           }
    {  170 }  CR( 6,  2,  AM, 0,  0,  0,  0),     {  Amplitude   }
    {  171 }  CR( 8,  0,  PR, 0,  0,  0,  0),     {  Period      }
    {  172 }  CR( 6,  2,  F0, 0,  1,  0,  0),     {  F0 (6 MSBs) }
    {  173 }  CR( 6,  2,  F1, 0,  1,  0,  0),     {  F1 (6 MSBs) }
    {  174 }  CR( 6,  2,  F2, 0,  1,  0,  0),     {  F2 (6 MSBs) }
                { SETMSB_3 only            }
    {  175 }  CR( 5,  0,  IA, 0,  0,  0,  0),     {  Ampl. Intr. }
    {  176 }  CR( 5,  0,  IP, 0,  0,  0,  0),     {  Per. Intr.  }
};
*)


Const
  SP0256_Df_Idx : Array[0..127] Of Integer =
(
    {  OPCODE 0000 }      -1, -1,     -1, -1,     -1, -1,     -1, -1,
    {  OPCODE 1000 }      -1, -1,     -1, -1,     -1, -1,     -1, -1,
    {  OPCODE 0100 }      17, 22,     17, 24,     25, 30,     25, 32,
    {  OPCODE 1100 }      83, 94,     129,142,    97, 108,    145,158,
    {  OPCODE 0010 }      83, 96,     129,144,    97, 110,    145,160,
    {  OPCODE 1010 }      73, 77,     74, 77,     78, 82,     79, 82,
    {  OPCODE 0110 }      33, 36,     34, 37,     38, 41,     39, 42,
    {  OPCODE 1110 }      127,128,    127,128,    127,128,    127,128,
    {  OPCODE 0001 }      1,  14,     1,  16,     1,  14,     1,  16,
    {  OPCODE 1001 }      45, 56,     45, 58,     59, 70,     59, 72,
    {  OPCODE 0101 }      161,166,    162,166,    169,174,    170,174,
    {  OPCODE 1101 }      111,116,    111,118,    119,124,    119,126,
    {  OPCODE 0011 }      161,168,    162,168,    169,176,    170,176,
    {  OPCODE 1011 }      -1, -1,     -1, -1,     -1, -1,     -1, -1,
    {  OPCODE 0111 }      -1, -1,     -1, -1,     -1, -1,     -1, -1,
    {  OPCODE 1111 }      0,  0,      0,  0,      0,  0,      0,  0
);

{ ======================================================================== }
{  BITREV       -- Bit-reverse a 32-bit number.                            }
{ ======================================================================== }
Function BitRev(Val: Cardinal): Cardinal;
Begin
  Val := ((Val And $FFFF0000) Shr 16) Or ((Val And $0000FFFF) Shl 16);
  Val := ((Val And $FF00FF00) Shr  8) Or ((Val And $00FF00FF) Shl  8);
  Val := ((Val And $F0F0F0F0) Shr  4) Or ((Val And $0F0F0F0F) Shl  4);
  Val := ((Val And $CCCCCCCC) Shr  2) Or ((Val And $33333333) Shl  2);
  Val := ((Val And $AAAAAAAA) Shr  1) Or ((Val And $55555555) Shl  1);
  BitRev := Val;
End; { BitRev }

{ ======================================================================== }
{  SP0256_GETB  -- Get up to 8 bits at the current PC.                     }
{ ======================================================================== }
Function SP0256_GetB(Var IVoice: TIVoice; Len: Word): Byte;
Const Data: LongInt = 0;
Var
  D0, D1 : LongInt;
  Idx0   : Word;
  Idx1   : Word;
  Page0  : Word;
  Page1  : Word;

Begin
  { -------------------------------------------------------------------- }
  {  Fetch data from the FIFO or from the MASK                           }
  { -------------------------------------------------------------------- }
  If IVoice.Fifo_Sel Then
  Begin
    D0 := IVoice.Fifo[(IVoice.Fifo_Tail    ) And 63];
    D1 := IVoice.Fifo[(IVoice.Fifo_Tail + 1) And 63];
    Data := ((D1 Shl 10) Or D0) Shr IVoice.Fifo_BitP;

    { ---------------------------------------------------------------- }
    {  Note the PC doesn't advance when we execute from FIFO.          }
    {  Just the FIFO's bit-pointer advances.   (That's not REALLY      }
    {  what happens, but that's roughly how it behaves.)               }
    { ---------------------------------------------------------------- }
    Inc(IVoice.Fifo_BitP,Len);
    If IVoice.Fifo_BitP >= 10 Then
    Begin
      Inc(IVoice.Fifo_Tail);
      Dec(IVoice.Fifo_BitP,10);
    End;
  End
  Else
  Begin
    { ---------------------------------------------------------------- }
    {  Figure out which ROMs are being fetched into, and grab two      }
    {  adjacent bytes.  The byte we're interested in is extracted      }
    {  from the appropriate bit-boundary between them.                 }
    { ---------------------------------------------------------------- }
    Idx0 := (IVoice.PC    ) Shr 3; Page0 := Idx0 Shr 12;
    Idx1 := (IVoice.PC + 8) Shr 3; Page1 := Idx1 Shr 12;

    Idx0 := Idx0 And $FFF;
    Idx1 := Idx1 And $FFF;

    D0 := 0;
    D1 := 0;

    If IVoice.Rom[Page0] <> Nil Then D0 := IVoice.Rom[Page0]^[Idx0];
    If IVoice.Rom[Page1] <> Nil Then D1 := IVoice.Rom[Page1]^[Idx1];

    Data := ((D1 Shl 8) Or D0) Shr (IVoice.PC And 7);

    Inc(IVoice.PC,Len);
  End;

  { -------------------------------------------------------------------- }
  {  Mask data to the requested length.                                  }
  { -------------------------------------------------------------------- }

  Data := Data And ((1 Shl Len) - 1);

  SP0256_GetB := Data And $FF;
End; { SP0256_GetB }

{ ======================================================================== }
{  SP0256_MICRO -- Emulate the microcontroller in the SP0256.  Executes    }
{                  instructions either until the repeat count <> 0 or      }
{                  the controller gets halted by a RTS to 0.               }
{ ======================================================================== }
Procedure SP0256_Micro(Var IV: TIVoice);
Var
  Immed4        : Byte;
  Opcode        : Byte;
  CR            : Word;
  I, J          : Integer;
  Idx0, Idx1    : Integer;
  Btrg          : Cardinal;

  Len, Shf, Delta, Field, Prm, ClrA, Clr5 : Word;
  Value  : ShortInt;
  OldVal : ShortInt;

Const
  Ctrl_Xfer : Boolean = False;
  _Repeat   : Integer = 0;

Label Continue,ContFor;
Begin
  { -------------------------------------------------------------------- }
  {  Only execute instructions while the filter is not busy.             }
  { -------------------------------------------------------------------- }
  While IV.Filt.Rpt <= 0 Do
  Begin
    { ---------------------------------------------------------------- }
    {  If the CPU is Halted, see if we have a new command pending      }
    {  in the Address LoaD buffer.                                     }
    { ---------------------------------------------------------------- }
    If (IV.Halted And (IV.LRQ = 0)) Then
    Begin
      IV.PC     := Word(IV.ALD) Or $8000;
      IV.Halted := False;
      IV.LRQ    := $8000;
      IV.ALD    := 0;
    End;

    { ---------------------------------------------------------------- }
    {  If we're still Halted, do nothing.                              }
    { ---------------------------------------------------------------- }
    If IV.Halted Then
    Begin
      IV.Filt.Rpt := 1;
      IV.LRQ      := $8000;
      IV.ALD      := 0;
      Exit;
    End;

    { ---------------------------------------------------------------- }
    {  Fetch the first 8 bits of the opcode, which are always in the   }
    {  same approximate format -- immed4 followed by opcode.           }
    { ---------------------------------------------------------------- }
    Immed4    := SP0256_GetB(IV, 4);
    Opcode    := SP0256_GetB(IV, 4);
    _Repeat   := 0;
    Ctrl_Xfer := False;

    { ---------------------------------------------------------------- }
    {  Handle the special cases for specific opcodes.                  }
    { ---------------------------------------------------------------- }
    Case Opcode Of
      { ------------------------------------------------------------ }
      {  OPCODE 0000:  RTS / SETPAGE                                 }
      { ------------------------------------------------------------ }
      $0:
      Begin
        { -------------------------------------------------------- }
        {  If immed4 <> 0, then this is a SETPAGE instruction.     }
        { -------------------------------------------------------- }
        If Immed4 <> 0 Then     { SETPAGE }
        Begin
          IV.Page := BitRev(Immed4) Shr 13;
        End
        Else
        { -------------------------------------------------------- }
        {  Otherwise, this is an RTS / HLT.                        }
        { -------------------------------------------------------- }
        Begin
          { ---------------------------------------------------- }
          {  Figure out our branch target.                       }
          { ---------------------------------------------------- }
          Btrg := IV.Stack;

          IV.Stack := 0;

          { ---------------------------------------------------- }
          {  If the branch target is zero, this is a HLT.        }
          {  Otherwise, it's an RTS, so set the PC.              }
          { ---------------------------------------------------- }
          If Btrg = 0 Then
          Begin
            IV.Halted := True;
            IV.PC     := 0;
            Ctrl_Xfer := True;
          End
          Else
          Begin
            IV.PC     := Btrg;
            Ctrl_Xfer := True;
          End;
        End;
      End;

      { ------------------------------------------------------------ }
      {  OPCODE 0111:  JMP          Jump to 12-bit/16-bit Abs Addr   }
      {  OPCODE 1011:  JSR          Jump to Subroutine               }
      { ------------------------------------------------------------ }
      $D,$E:
      Begin

        { -------------------------------------------------------- }
        {  Figure out our branch target.                           }
        { -------------------------------------------------------- }
        Btrg := IV.Page                             Or
                (BitRev(Immed4)             Shr 17) Or
                (BitRev(SP0256_GetB(IV, 8)) Shr 21);
        Ctrl_Xfer := True;

        { -------------------------------------------------------- }
        {  If this is a JSR, push our return address on the        }
        {  stack.  Make sure it's byte aligned.                    }
        { -------------------------------------------------------- }
        If Opcode = $D Then IV.Stack := (IV.PC + 7) And $FFF8;

        { -------------------------------------------------------- }
        {  Jump to the new location!                               }
        { -------------------------------------------------------- }
        IV.PC := Btrg;
      End;

      { ------------------------------------------------------------ }
      {  OPCODE 1000:  SETMODE      Set the Mode and Repeat MSBs     }
      { ------------------------------------------------------------ }
      $1:
      Begin
        IV.Mode := ((Immed4 And 8) Shr 2) Or (Immed4 And 4) Or
                   ((Immed4 And 3) Shl 4);
      End;

      { ------------------------------------------------------------ }
      {  OPCODE 0001:  LOADALL     Load All Parameters               }
      {  OPCODE 0010:  LOAD_2      Load Per, Ampl, Coefs, Interp.    }
      {  OPCODE 0011:  SETMSB_3    Load Pitch, Ampl, MSBs, And Intrp }
      {  OPCODE 0100:  LOAD_4      Load Pitch, Ampl, Coeffs          }
      {  OPCODE 0101:  SETMSB_5    Load Pitch, Ampl, and Coeff MSBs  }
      {  OPCODE 0110:  SETMSB_6    Load Ampl, and Coeff MSBs.        }
      {  OPCODE 1001:  DELTA_9     Delta update Ampl, Pitch, Coeffs  }
      {  OPCODE 1010:  SETMSB_A    Load Ampl and MSBs of 3 Coeffs    }
      {  OPCODE 1100:  LOAD_C      Load Pitch, Ampl, Coeffs          }
      {  OPCODE 1101:  DELTA_D     Delta update Ampl, Pitch, Coeffs  }
      {  OPCODE 1110:  LOAD_E      Load Pitch, Amplitude             }
      {  OPCODE 1111:  PAUSE       Silent pause                      }
      { ------------------------------------------------------------ }
    Else
      _Repeat := Immed4 Or (IV.Mode And $30);
    End; { Case }
    If Opcode <> 1 Then IV.Mode := IV.Mode And $F;

    { ---------------------------------------------------------------- }
    {  If this was a control transfer, handle setting "Fifo_Sel"       }
    {  and all that ugliness.                                          }
    { ---------------------------------------------------------------- }
    If Ctrl_Xfer Then
    Begin

      { ------------------------------------------------------------ }
      {  Set our "FIFO Selected" flag based on whether we're going   }
      {  to the FIFO's address.                                      }
      { ------------------------------------------------------------ }
      IV.Fifo_Sel := (IV.PC = Fifo_ADDR);

      { ------------------------------------------------------------ }
      {  Control transfers to the FIFO cause it to discard the       }
      {  partial decle that's at the front of the FIFO.              }
      { ------------------------------------------------------------ }
      If IV.Fifo_Sel And (IV.Fifo_BitP <> 0) Then
      Begin

        { Discard partially-read decle. }
        If (IV.Fifo_Tail < IV.Fifo_Head) Then Inc(IV.Fifo_Tail);
        IV.Fifo_BitP := 0;
      End;
      Goto Continue;
    End;

    { ---------------------------------------------------------------- }
    {  Otherwise, if we have a repeat count, then go grab the data     }
    {  block and feed it to the filter.                                }
    { ---------------------------------------------------------------- }
    If _Repeat = 0 Then Goto Continue;
    IV.Filt.Rpt := _Repeat + 1;
    I    := (Opcode Shl 3) Or (IV.Mode And 6);
    Idx0 := SP0256_Df_Idx[I];
    Idx1 := SP0256_Df_Idx[I + 1];

    { ---------------------------------------------------------------- }
    {  Step through control words in the description for data block.   }
    { ---------------------------------------------------------------- }
    For I := Idx0 To Idx1 Do
    Begin

      { ------------------------------------------------------------ }
      {  Get the control word and pull out some important fields.    }
      { ------------------------------------------------------------ }
      CR := SP0256_DataFmt[I];

      Len   := CR_LEN(CR);
      Shf   := CR_SHF(CR);
      Prm   := CR_PRM(CR);
      ClrA  := CR And CR_CLRA;
      Clr5  := CR And CR_CLR5;
      Delta := CR And CR_DELTA;
      Field := CR And CR_FIELD;

      { ------------------------------------------------------------ }
      {  Clear any registers that were requested to be cleared.      }
      { ------------------------------------------------------------ }
      If ClrA <> 0 Then For J := 0 To 15 Do IV.Filt.R[J] := 0;

      If Clr5 <> 0 Then
      Begin
        IV.Filt.R[Word(B5)] := 0;
        IV.Filt.R[Word(F5)] := 0;
      End;

      { ------------------------------------------------------------ }
      {  If this entry has a bitfield with it, grab the bitfield.    }
      { ------------------------------------------------------------ }
      If Len = 0 Then Goto ContFor;
      Value := ShortInt(SP0256_GetB(IV, Len));

      { ------------------------------------------------------------ }
      {  Sign extend if this is a delta update.                      }
      { ------------------------------------------------------------ }
      If Delta <> 0 Then { Sign extend }
      Begin
        If (Value And (1 Shl (Len - 1))) <> 0 Then
         Value := Value Or ($FF Shl Len);
      End;

      { ------------------------------------------------------------ }
      {  Shift the value to the appropriate precision.               }
      { ------------------------------------------------------------ }
      If Shf <> 0 Then Value := ShortInt(Value Shl Shf);

      { ------------------------------------------------------------ }
      {  If this is a field-replace, insert the field.               }
      { ------------------------------------------------------------ }
      If Field <> 0 Then
      Begin
        IV.Filt.R[Prm] := IV.Filt.R[Prm] And ($FF Xor ($FF Shl Shf)); { Clear the old bits.     }
        IV.Filt.R[Prm] := IV.Filt.R[Prm] Or Byte(Value);              { Merge in the new bits.  }
        Goto ContFor;
      End;

      { ------------------------------------------------------------ }
      {  If this is a delta update, add to the appropriate field.    }
      { ------------------------------------------------------------ }
      If Delta <> 0 Then
      Begin
        OldVal := ShortInt(IV.Filt.R[Prm]);
        Inc(OldVal,Value);
        IV.Filt.R[Prm] := Byte(OldVal);
        Goto ContFor;
      End;

      { ------------------------------------------------------------ }
      {  Otherwise, just write the new value.                        }
      { ------------------------------------------------------------ }
      IV.Filt.R[Prm] := Byte(Value);
ContFor:
    End; { For I }

    { ---------------------------------------------------------------- }
    {  Special case:  Set PAUSE's equivalent period.                   }
    { ---------------------------------------------------------------- }
    If Opcode = $F Then IV.Filt.R[1] := PER_PAUSE;

    { ---------------------------------------------------------------- }
    {  Now that we've updated the registers, go decode them.           }
    { ---------------------------------------------------------------- }
    LPC12_RegDec(IV.Filt);

    { ---------------------------------------------------------------- }
    {  Break out since we now have a repeat count.                     }
    { ---------------------------------------------------------------- }
    Exit;
Continue:
  End; { While }
End; { SP0256_Micro }

{ ======================================================================== }
{  IVoice_TK    -- Where the magic happens.  Generate voice data for       }
{                  our good friend, the IntellIVoice.                      }
{ ======================================================================== }
Function IVoice_Tk(Var IVoice: TIVoice; Len: Word): LongInt;
Var
  Did_Samp  : Word;
  Index     : Word;

Begin
  { -------------------------------------------------------------------- }
  {  Iterate the sound engine.                                           }
  { -------------------------------------------------------------------- }
  Index     := 0;
  IVoice_Tk := Len;
  While Len > 0 Do
  Begin
    { ------------------------------------------------------------ }
    {  If our repeat count expired, emulate the microcontroller.   }
    { ------------------------------------------------------------ }
    If IVoice.Filt.Rpt <= 0 Then SP0256_Micro(IVoice);
    Did_Samp := LPC12_Update(IVoice.Filt, Len, Index, IVoice.Rate);
    Dec(Len,Did_Samp);
    Inc(Index,Did_Samp);
  End; // While
End; { IVoice_Tk }

{ ======================================================================== }
{  IVoice_RD    -- Handle reads from the IntellIVoice.                     }
{ ======================================================================== }
Function IVoice_Rd(Var IVoice: TIVoice; Addr: Word): Word;
Begin

  { -------------------------------------------------------------------- }
  {  Address $80 returns the SP0256 LRQ status on bit 15.                }
  { -------------------------------------------------------------------- }
  If Addr = 0 Then
  Begin
    IVoice_Rd := IVoice.LRQ;
  End
  Else

  { -------------------------------------------------------------------- }
  {  Address $81 returns the SPB640 FIFO full status on bit 15.          }
  { -------------------------------------------------------------------- }
  If Addr = 1 Then
  Begin
    If (IVoice.Fifo_Head - IVoice.Fifo_Tail) >= 64
     Then IVoice_Rd := $8000
     Else IVoice_Rd := 0;
  End
  Else

  { -------------------------------------------------------------------- }
  {  Just return 255 for all other addresses in our range.               }
  { -------------------------------------------------------------------- }
   IVoice_Rd := $00FF;
End; { IVoice_Rd }

{ ======================================================================== }
{  IVoice_WR    -- Handle writes to the IntellIVoice.                      }
{ ======================================================================== }
Procedure IVoice_Wr(Var IVoice: TIVoice; Addr,Data: Word);
Begin

  { -------------------------------------------------------------------- }
  {  Ignore writes outside $80, $81.                                     }
  { -------------------------------------------------------------------- }
  If Addr > 1 Then Exit;

  { -------------------------------------------------------------------- }
  {  Address $80 is for Address Loads (essentially speech commands).     }
  { -------------------------------------------------------------------- }
  If Addr = 0 Then
  Begin
    { ---------------------------------------------------------------- }
    {  Drop writes to the ALD register if we're busy.                  }
    { ---------------------------------------------------------------- }
    If IVoice.LRQ = 0 Then Exit;

    { ---------------------------------------------------------------- }
    {  Set LRQ to 1 and load the 8 LSBs of the data into the ALR reg.  }
    {  We take the command address, and multiply by 2 bytes to get     }
    {  the new PC address.                                             }
    { ---------------------------------------------------------------- }
    IVoice.LRQ := 0;
    IVoice.ALD := ($FF And Data) Shl 4;
  End
  Else

  { -------------------------------------------------------------------- }
  {  Address $81 is for FIFOing up decles.  The FIFO is 64 decles       }
  {  long.  The Head pointer points to where we insert new decles and    }
  {  the Tail pointer is where we pull them from.                        }
  { -------------------------------------------------------------------- }
  If Addr = 1 Then
  Begin
    { ---------------------------------------------------------------- }
    {  If Bit 10 is set, just reset the FIFO and SP0256.               }
    { ---------------------------------------------------------------- }
    If (Data And $400) <> 0 Then
    Begin
      IVoice.Fifo_Head := 0;
      IVoice.Fifo_Tail := 0;
      IVoice.Fifo_BitP := 0;
      FillChar(IVoice.Filt,SizeOf(IVoice.Filt),#0);
      IVoice.Halted   := True;
      IVoice.Filt.Rpt := -1;
      IVoice.Filt.Rng := 1;
      IVoice.LRQ      := $8000;
      IVoice.PC       := $0000;
      IVoice.Stack    := $0000;
    End
    Else
    Begin
      { -------------------------------------------------------------- }
      {  If the FIFO is full, drop the data.                           }
      { -------------------------------------------------------------- }
      If (IVoice.Fifo_Head - IVoice.Fifo_Tail) < 64 Then
      Begin

        { ------------------------------------------------------------ }
        {  FIFO up the lower 10 bits of the data.                      }
        { ------------------------------------------------------------ }
        IVoice.Fifo[IVoice.Fifo_Head And 63] := Data And $3FF;
        Inc(IVoice.Fifo_Head);
      End;
    End;
  End;
End; { IVoice_Wr }

{ ======================================================================== }
{  IVoice_INIT  -- Makes a new IntellIVoice                                }
{ ======================================================================== }
Function IVoice_Init(Var IVoice : TIVoice;        { Structure to initialize. }
                         Rate   : Word): Boolean; { Desired sample rate.     }
Begin

  { -------------------------------------------------------------------- }
  {  First, lets zero out the structure to be safe.                      }
  { -------------------------------------------------------------------- }

  FillChar(IVoice,SizeOf(IVoice),#0);

  { -------------------------------------------------------------------- }
  {  Sanity checks.                                                      }
  { -------------------------------------------------------------------- }

  If (Rate < 10000) Or (Rate > 48000) Then
  Begin
    IVoice_Init := False;
    Exit;
  End;

  { -------------------------------------------------------------------- }
  {  Configure our internal variables.                                   }
  { -------------------------------------------------------------------- }
  IVoice.Rom[1]   := @Mask;
  IVoice.Rate     := Rate;
  IVoice.Filt.Rng := 1;

  { -------------------------------------------------------------------- }
  {  Set up the microcontroller's initial state.                         }
  { -------------------------------------------------------------------- }
  IVoice.Halted   := True;
  IVoice.Filt.Rpt := -1;
  IVoice.LRQ      := $8000;
  IVoice.Page     := $8000;
  FillChar(VoiceBuf,SizeOf(VoiceBuf),#0);
  IVoice_Init := True;
  Samp_Count  := Rate;
End; { IVoice_Init }

Procedure MakeVestigialROM;
Var I: Integer;
Begin
  FillChar(Mask,SizeOf(Mask),#0);        { Clear the whole array }

  { Initialize the jump table. We will have only two code areas. }

  Mask[0] := $E0;             { For FIFO loading: Jump to 1060h }
  Mask[1] := $06;
  For I := 1 To 47 Do
  Begin
    Mask[I * 2]     := $E0;   { Pause-and-return: Jump to 1064h }
    Mask[I * 2 + 1] := $26;
  End; { For I }

  { Input SP0256 code for address 0: load-from-FIFO }

  Mask[$60] := $F1;   { Short pause }
  Mask[$61] := $E1;   { Jump to 1800h }
  Mask[$62] := $00;
  Mask[$63] := $00;   { Return }

  { Input SP0256 code for all other addresses (1-47): pause-and-return }

  Mask[$64] := $F7;   { Short pause }
  Mask[$65] := $00;   { Return      }
End; { MakeVestigialROM }

{ ======================================================================== }
{  This program is free software; you can redistribute it and/or modify    }
{  it under the terms of the GNU General Public License as published by    }
{  the Free Software Foundation; either version 2 of the License, or       }
{  (at your option) any later version.                                     }
{                                                                          }
{  This program is distributed in the hope that it will be useful,         }
{  but WITHOUT ANY WARRANTY; without even the implied warranty of          }
{  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       }
{  General Public License for more details.                                }
{                                                                          }
{  You should have received a copy of the GNU General Public License       }
{  along with this program; if not, write to the Free Software             }
{  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               }
{ ======================================================================== }
{                 Copyright (c) 1998-2000, Joseph Zbiciak                  }
{ ======================================================================== }

Initialization
  { Assume we don't have an Intellivoice ROM available. Initialize the array }
  { and put a little SP0256 code into it so that the emulator can at least   }
  { function.                                                                }

  MakeVestigialROM;
End.



